@@ -7,7 +7,7 @@
  controlling Pixelblaze LED controllers.  Requires Python 3 and the websocket-client
  module.
 
- Copyright 2020 JEM (ZRanger1)
+ Copyright 2020-2022 JEM (ZRanger1)
 
  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
@@ -26,11 +26,13 @@
  THE SOFTWARE.
 
  Version  Date         Author Comment
- v0.0.1   11/20/2020   JEM(ZRanger1)    Created
- v0.0.2   12/1/2020    JEM(ZRanger1)    Name change + color control methods
- v0.9.0   12/6/2020    JEM(ZRanger1)    Added PixelblazeEnumerator class
- v0.9.1   12/16/2020   JEM(ZRanger1)    Support for pypi upload
- v0.9.2   01/16/2020   JEM(ZRanger1)    Updated Pixelblaze sequencer support
+ v0.0.1   11/20/2020   JEM(ZRanger1)  Created
+ v0.0.2   12/1/2020    "              Name change + color control methods
+ v0.9.0   12/6/2020    "              Added PixelblazeEnumerator class
+ v0.9.1   12/16/2020   "              Support for pypi upload
+ v0.9.2   01/16/2021   "              Updated Pixelblaze sequencer support
+ v0.9.3   04/13/2021   "              waitForEmptyQueue() return now agrees w/docs
+ v0.9.4   02/04/2022   "              Added setPixelcount(),pause(), unpause(), pattern cache
 """
 import websocket
 import socket
@@ -39,7 +41,7 @@ import time
 import struct
 import threading
 
-__version__ = "0.9.2"
+__version__ = "0.9.4"
 
 class Pixelblaze:
     """
@@ -54,12 +56,17 @@ class Pixelblaze:
     flash_save_enabled = False
     default_recv_timeout = 1
     ipAddr = None
-
+    
+    cacheRefreshTime = 0
+    cacheRefreshInterval = 1000; # milliseconds used internally
+    patternCache = None
+    
     def __init__(self, addr):
         """
         Create and open Pixelblaze object. Takes the Pixelblaze's IPv4 address in the
         usual 12 digit numeric form (for example, 192.168.1.xxx) 
         """
+        self.setCacheRefreshTime(600)  # seconds used in public api
         self.open(addr)
 
     def open(self, addr):
@@ -69,7 +76,7 @@ class Pixelblaze:
         explicitly call open to connect unless the websocket has been closed by the
         user or by the Pixelblaze.
         """
-        if (self.connected is False):
+        if self.connected is False:
             uri = "ws://" + addr + ":81"
             self.ws = websocket.create_connection(uri, sockopt=((socket.SOL_SOCKET, socket.SO_REUSEADDR, 1),
                                                                 (socket.IPPROTO_TCP, socket.TCP_NODELAY, 1),))
@@ -79,7 +86,7 @@ class Pixelblaze:
 
     def close(self):
         """Close websocket connection"""
-        if (self.connected is True):
+        if self.connected is True:
             self.ws.close()
             self.connected = False
 
@@ -114,20 +121,24 @@ class Pixelblaze:
         packets before sending requests for data w/send_string(). We do not call it
         before simply sending commands because it has a small performance cost.
         
-        This is one of the treacherously "clever" things done to make pyblaze
+        This is one of the treacherously "clever" things done to make pixelblaze-client
         work as a synchronous API when the Pixelblaze may be sending out unexpected
         packets or talking to multiple devices.  We do some extra work to make sure
         we're only receiving the packets we want.
         """
 
-        self.ws.settimeout(0.1)  # set very short timeout
-
+        # set very short timeout.
+        self.ws.settimeout(0.1)  
+            
+        # eat packets until we get a timeout exception on recv(), indicating 
+        # that there are no more pending packets
         try:
-            while (True):
-                self.ws.recv()
+            while True:
+                self.ws.recv()                         
         except websocket._exceptions.WebSocketTimeoutException:
-            self.ws.settimeout(self.default_recv_timeout)  # restore normal timeout
-            return  # if we get a timeout, there are no more pending packets
+            # restore normal timeout when done
+            self.ws.settimeout(self.default_recv_timeout)  
+        return  
 
     def ws_recv(self, wantBinary=False):
         """
@@ -136,11 +147,11 @@ class Pixelblaze:
         """
         result = None
         try:
-            while (True):  # loop until we hit timeout or have the packet we want
+            while True:  # loop until we hit timeout or have the packet we want
                 result = self.ws.recv()
-                if ((wantBinary is False) and (type(result) is str)):  # JSON string
+                if (wantBinary is False) and (type(result) is str):  # JSON string
                     break
-                elif ((wantBinary is True) and (result[0] == 0x07)):  # binary pattern list packet
+                elif (wantBinary is True) and (result[0] == 0x07):  # binary pattern list packet
                     break
                 else:
                     result = None  # unhandled binary - ignore
@@ -162,18 +173,26 @@ class Pixelblaze:
         """
         Wait until the Pixelblaze's websocket message queue is empty, or until
         timeout_ms milliseconds have elapsed.  Returns True if an empty queue
-        acknowldgement was received, False if timeout or error occurs.
+        acknowledgement was received, False otherwise.  Throws an exception
+        if the socket is disconnected.
         """
         self.ws_flush()
         self.ws.settimeout(timeout_ms / 1000)
-        self.send_string('{"ping": true}')
-        result = self.ws.recv()
-        self.ws.settimeout(self.default_recv_timeout)
-
-        return True if ((result is not None) and (result.startswith('{"ack"'))) else False
+        try:
+            self.send_string('{"ping": true}')
+            result = self.ws.recv()
+            self.ws.settimeout(self.default_recv_timeout)
+            
+            return True if ((result is not None) and (result.startswith('{"ack"'))) else False       
+        except websocket._exceptions.WebSocketTimeoutException:
+            self.ws.settimeout(self.default_recv_timeout)
+          
+        return False
 
     def getVars(self):
-        """Returns JSON object containing all vars exported from the active pattern"""
+        """
+        Returns JSON object containing all vars exported from the active pattern
+        """
         self.ws_flush()  # make sure there are no pending packets    
         self.send_string('{"getVars": true}')
         return json.loads(self.ws.recv()).get('vars')
@@ -200,7 +219,7 @@ class Pixelblaze:
         False otherwise.
         """
         val = self.getVars()
-        if (val is None):
+        if val is None:
             return False
 
         return True if var_name in val else False
@@ -217,12 +236,11 @@ class Pixelblaze:
         """Utility Method: Returns a pattern ID if passed either a valid ID or a text name"""
         patterns = self.getPatternList()
 
-        if (patterns.get(pid) is None):
+        if patterns.get(pid) is None:
             pid = self._id_from_name(patterns, pid)
 
         return pid
-
-    def setActivePatternId(self, pid):
+    def setActivePatternId(self, pid,  saveFlash=False):
         """
         Sets the active pattern by pattern ID, without the name lookup option
         supported by setActivePattern().  This method is faster and more
@@ -232,18 +250,22 @@ class Pixelblaze:
         It does not validate the input id, or determine if the pattern is
         available on the Pixelblaze.
         """
-        self.send_string('{"activeProgramId" : "%s"}' % pid)
+        saveStr = self.__get_save_string(saveFlash)
+        self.send_string('{"activeProgramId" : "%s" %s}' % (pid, saveStr))
+        
+        # give the Pixelblaze a moment to load the pattern
+        self.waitForEmptyQueue(1000)         
 
-    def setActivePattern(self, pid):
+    def setActivePattern(self, pid, saveFlash=False):
         """Sets the currently running pattern, using either an ID or a text name"""
         p = self._get_pattern_id(pid)
 
-        if (p is not None):
-            self.setActivePatternId(p)
+        if p is not None:
+            self.setActivePatternId(p, saveFlash)
 
     def getActivePattern(self):
         """
-        Returns the ID the pattern currently running on
+        Returns the ID of the pattern currently running on
         the Pixelblaze if available.  Otherwise returns an empty dictionary
         object
         """
@@ -253,10 +275,47 @@ class Pixelblaze:
         except:
             return {}
 
-    def setBrightness(self, n):
-        """Set the Pixelblaze's global brightness.  Valid range is 0-1"""
+    def setBrightness(self, n, saveFlash=False):
+        """
+        Set the Pixelblaze's global brightness.  Valid range is 0-1
+        
+        The optional saveFlash parameter controls whether or not the
+        new brightness value is saved. By default, saveFlash is False, so
+        brightness values set with this method will not persist
+        through reboots.
+        
+        To reduce wear on the Pixelblaze's flash memory, the saveFlash parameter is ignored
+        by default.  See documentation for _enable_flash_save() for
+        more information.        
+        """
         n = max(0, min(n, 1))  # clamp to proper 0-1 range
-        self.send_string('{"brightness" : %f}' % n)
+        
+        saveStr = self.__get_save_string(saveFlash)
+        self.send_string('{"brightness" : %f %s}' % (n, saveStr))
+        
+    def setPixelCount(self, nPixels):
+        """
+        Sets the number of LEDs attached to the Pixelblaze. Does
+        not change the current pixel map. CAUTION: Recommended for
+        advanced users only. 
+        """       
+        # slow and careful, pausing rendering before we reset
+        # the pixel count, to avoid disruption.
+        self.pause();
+        self.ws_flush()
+        self.send_string('{"pixelCount":%d}' % nPixels)        
+        time.sleep(0.5)
+        self.unpause();
+        
+    def getPixelCount(self):
+        """
+        Returns the number of LEDs attached to the Pixelblaze.
+        """       
+        hw = self.getHardwareConfig()
+        try:
+            return hw['pixelCount']
+        except:
+            return {}       
 
     def setSequenceTimer(self, n):
         """
@@ -265,7 +324,7 @@ class Pixelblaze:
         """
         self.send_string('{"sequenceTimer" : %d}' % n)
 
-    def startSequencer(self,mode = 1):
+    def startSequencer(self, mode=1):
         """
         Enable and start the Pixelblaze's internal sequencer. The mode parameters
         can be 1 - shuffle all patterns, or 2 - playlist mode.  The playlist
@@ -301,7 +360,7 @@ class Pixelblaze:
         hw = dict()
 
         p1 = self.ws_recv()
-        while (p1 is not None):
+        while p1 is not None:
             p2 = json.loads(p1)
             hw = {**hw, **p2}
             p1 = self.ws_recv()
@@ -340,18 +399,21 @@ class Pixelblaze:
         # and get stored values for that program
         else:
             pattern = self._get_pattern_id(pattern)
-            if (pattern is None):
+            if pattern is None:
                 return None
 
             self.send_string('{"getControls": "%s"}' % pattern)
             ctl = json.loads(self.ws.recv())
-
+            
         # extract controls and their values
-        if (len(ctl.get('controls')) > 0):
-            x = next(iter(ctl['controls']))
-            return ctl['controls'][x]
-        else:
-            return {}
+        try:
+            if len(ctl.get('controls')) > 0:
+                x = next(iter(ctl['controls']))
+                return ctl['controls'][x]
+            else:
+                return {}
+        except:
+            return None
 
     def setControls(self, json_ctl, saveFlash=False):
         """
@@ -411,7 +473,7 @@ class Pixelblaze:
         argument is not specified, check the currently running pattern
         """
         controls = self.getControls(pattern)
-        if (controls is None):
+        if controls is None:
             return None
 
         # check for hsvPicker        
@@ -431,7 +493,7 @@ class Pixelblaze:
         argument is not specified, checks in the currently running pattern
         """
         result = self.getColorControlNames(pattern)
-        if (result is None):
+        if result is None:
             return result
         else:
             return result[0]
@@ -451,33 +513,87 @@ class Pixelblaze:
         saveStr = self.__get_save_string(saveFlash)
         self.ws_flush()
         self.send_string('{"dataSpeed":%d %s}' % (speed, saveStr))
-
-    def getPatternList(self):
+        
+    def pause(self):
+        """
+        Pause rendering. Lasts until unpause() is called or
+        the Pixelblaze is reset.
+        CAUTION: For advanced users only.  If you don't know
+        exactly why you want to do this, DON'T DO IT.
+        """
+        self.ws_flush()
+        self.send_string('{"pause" : true}')
+        
+    def unpause(self):
+        """
+        Resume rendering if halted by pause(). No effect otherwise.
+        CAUTION: For advanced users only.  If you don't know
+        exactly why you want to do this, DON'T DO IT.
+        """
+        self.ws_flush()
+        self.send_string('{"pause" : false}')
+        
+        
+    def getPatternList(self,refresh = False):
         """
         Returns a dictionary containing the unique ID and the text name of all
-        saved patterns on the Pixelblaze
+        saved patterns on the Pixelblaze. Normally reads from the cached pattern
+        list, which is refreshed every 10 minutes by default.
+        
+        To force a cache refresh, set the optional "refresh" parameter to True
+        
+        To change the cache refresh interval, call setCacheRefreshTime(seconds)
         """
-        patterns = dict()
+        if refresh is True or ((self._time_in_millis() - self.cacheRefreshTime) > self.cacheRefreshInterval):
+            self._refreshPatternCache()
+            self.cacheRefreshTime = self._time_in_millis()
+        
+        return self.patternCache
+    
+    def _clamp(self,n, smallest, largest):
+        """
+        Utility Method: Why doesn't Python have clamp()?
+        """ 
+        return max(smallest, min(n, largest))
+  
+    def _time_in_millis(self):
+        """
+        Utility Method: Returns current time in milliseconds
+        """
+        return int(round(time.time() * 1000))
+    
+    def setCacheRefreshTime(self,seconds):
+        """
+        Set the interval, in seconds, at which the pattern cache is cleared and
+        a new pattern list is loaded from the Pixelblaze.  Default is 600 (10 minutes)
+        """        
+        # a million seconds is about 277 hours
+        # or about 11.5 days.  Probably long enough.
+        self.cacheRefreshInterval = self._clamp(seconds,0,1000000) * 1000;      
+
+    def _refreshPatternCache(self):
+        """
+        Reads a dictionary containing the unique ID and the text name of all
+        saved patterns on the Pixelblaze into the pattern cache.
+        """
+        self.patternCache = dict()
         self.ws_flush()  # make sure there are no pending packets    
         self.send_string("{ \"listPrograms\" : true }")
 
         frame = self.ws_recv(True)
-        while (frame is not None):
+        while frame is not None:
             listFrame = frame[2:].decode("utf-8")
             listFrame = listFrame.split("\n")
             listFrame = [m.split("\t") for m in listFrame]
 
             for pat in listFrame:
-                if (len(pat) == 2):
-                    patterns[pat[0]] = pat[1]
+                if len(pat) == 2:
+                    self.patternCache[pat[0]] = pat[1]
 
-            if (frame[1] & 0x04):
+            if frame[1] & 0x04:
                 break
             frame = self.ws_recv(True)
 
-        return patterns
-
-
 class PixelblazeEnumerator:
     """
     Listens on a network to detect available Pixelblazes, which the user can then list
@@ -523,17 +639,17 @@ class PixelblazeEnumerator:
         packet, returning a 3 element list which contains
         (packet_type, sender_id, sender_time)
         """
-        return struct.unpack("LLL", data)
+        return struct.unpack("<LLL", data)
 
     def _pack_timesync(self, now, sender_id, sender_time):
         """
         Utility Method: Builds a Pixelblaze timesync packet from
         supplied data.
         """
-        return struct.pack("LLLLL", self.TIMESYNC_PACKET, self.SYNC_ID,
+        return struct.pack("<LLLLL", self.TIMESYNC_PACKET, self.SYNC_ID,
                            now, sender_id, sender_time)
     
-    def _set_timesync_id(self,id):
+    def _set_timesync_id(self, id):
         """Utility Method:  Sets the PixelblazeEnumerator object's network
            id for time synchronization. At the moment, any 32 bit value will
            do, and calling this method does (almost) nothing.  In the
@@ -591,7 +707,7 @@ class PixelblazeEnumerator:
         """
         Stop listening for datagrams, terminate listener thread and close socket.
         """
-        if (self.listener is None):
+        if self.listener is None:
             return
         else:
             self.isRunning = False
@@ -601,12 +717,12 @@ class PixelblazeEnumerator:
             self.threadObj = None
             self.listener = None
 
-    def _send_timesync(self,now,sender_id,sender_time,addr):
+    def _send_timesync(self, now, sender_id, sender_time, addr):
         """
         Utility Method: Composes and sends a timesync packet to a single Pixelblaze
         """
         try:
-            self.listener.sendto(self._pack_timesync(now,sender_id,sender_time),addr)
+            self.listener.sendto(self._pack_timesync(now, sender_id, sender_time), addr)
 
         except socket.error as e:
             print(e)
@@ -620,37 +736,36 @@ class PixelblazeEnumerator:
         while self.isRunning:
             data, addr = self.listener.recvfrom(1024)
             now = self._time_in_millis()
+            
+            # check the list periodically,and remove devices we haven't seen in a while
+            if (now - self.listTimeoutCheck) >= self.LIST_CHECK_INTERVAL:
+                newlist = dict()
+
+                for dev, record in self.devices.items():
+                    if (now - record["timestamp"]) <= self.DEVICE_TIMEOUT:
+                        newlist[dev] = record
+
+                self.devices = newlist
+                self.listTimeoutCheck = now                        
 
             # when we receive a beacon packet from a Pixelblaze,
             # update device record and timestamp in our device list
             pkt = self._unpack_beacon(data)
-            if (pkt[0] == self.BEACON_PACKET):
-                #add pixelblaze to list of devices
+            if pkt[0] == self.BEACON_PACKET:
+                # add pixelblaze to list of devices
                 self.devices[pkt[1]] = {"address": addr,
                                         "timestamp": now,
                                         "sender_id": pkt[1],
                                         "sender_time": pkt[2]}
                 
                 # immediately send timesync if enabled
-                if self.autoSync:
-                    self._send_timesync(now,pkt[1],pkt[2],addr)
-
-                # remove devices we haven't seen in a while
-                if ((now - self.listTimeoutCheck) >= self.LIST_CHECK_INTERVAL):
-                    newlist = dict()
-
-                    for dev, record in self.devices.items():
-                        if ((now - record["timestamp"]) <= self.DEVICE_TIMEOUT):
-                            newlist[dev] = record
-
-                    self.devices = newlist
-                    self.listTimeoutCheck = now
+                if self.autoSync: # send
+                    self._send_timesync(now, pkt[1], pkt[2], addr)
                     
-            elif (pkt[0] == self.TIMESYNC_PACKET):   # always defer to other time sources
+            elif pkt[0] == self.TIMESYNC_PACKET:   # always defer to other time sources
                 if self.autoSync:
                     self.disableTimesync()
 
-
     def getPixelblazeList(self):
         dev = []
         for record in self.devices.values():
